// Copyright (c) Team CharLS.
// SPDX-License-Identifier: BSD-3-Clause

#pragma once

void performance_tests(int loop_count);
void decode_performance_tests(int loop_count);
void encode_performance_tests(int loop_count);
void test_large_image_performance_rgb8(int loop_count);
